/* Instructables: ESP32 No-Sleep-In Wi-Fi Alarm Clock in 3D Printed Enclosure
    by JD_K
    Link: https://www.instructables.com/ESP32-No-Sleep-In-Wi-Fi-Alarm-Clock-in-3D-Printed-/  
    To see a YouTube demo: https://youtu.be/dDLdVJgyW2U
*/

#include "WiFi.h"
#include "ESPAsyncWebServer.h"
#include "SPIFFS.h"

// ********** Defining the names of variables that we will use later **********

//Variables for the connections
const int buzzerPin = 32;
const int buttonPin = 26;
const int LEDpin = 25;
const int LDRpin = 35;
const byte clockPin = 21;
const byte latchPin = 19;
const byte dataPin = 18;
const byte dailyModePin = 33 ;

// Variables for Time and the alarm

tm localTime;
unsigned long prevLEDtime;
const unsigned long buzzerOverflow = 72E5;  //72E5 = (2hour*60min/hour*60sec/min*1000msec/sec);
unsigned long buzzerStartMillis;
int alarmHour;
int alarmMinute;
int snoozeHour;
int snoozeMinute;
bool buzzerFlag = 0;       // Used to signal if the buzzer should be sounding
bool toggleTone = 1;       // The buzzer will have two ones, this boolean used to trach which tone is next
unsigned long toggleTime;  // variable to be used to track when to change the tone of the buzzer
long toneDuration = 2000;
bool valuesUpdated = 0;
int DaY;
bool dailyMode = 0;

// Variables for displaying the time
volatile byte timeValues[] = {0, 0, 0, 0};
int delayTime = 3; // milliseconds delay for each digit
byte indexDigit = 0; // to cycle the digits one at a time should go from 1 to 4

// Variables for the light dependent resistor
int ambientLight;
int lightThreshold;
const int ambientTooBright = 2000;  // if the ambient light is too high already (sleeping in the day, or with light on already), it won't get much brighter
bool ambientChecked = false;        // when true, no need to check light again

// Variables for the RTOS multitasking
TaskHandle_t displayTimeTask0;
SemaphoreHandle_t baton;

enum listOfStates {
  noAlarm,        // Time is running but no alarm will be active in this state
  alarmSet,       // Alarm is set, but not sounding. This state will allow for a daily alarm and will check if the Alarm hour/min equal current hour/min
  alarmSounding,  // Buzzer triggered because: Alarm hour/min equal current hour/min. Ambient light level is sampled
  firstSnooze,    // Alarm sounded already but the person has snoozed first time.
  overSleeping,   // subsequent snoozing is oversleeping
  alarmOff,       // This temporary state for the alarm being turned off right away. Alarm hour/min will still equal current hour/min, after 1 min, will change to alarmSet
  // each of these is now like a number, with noAlarm==0, alarmSet==1, etc.
};

listOfStates alarmState = noAlarm;

// Replace with your network credentials
const char* ssid = "NetWorkName";
const char* password = "PassWord";


// Setting up real-time clock
const char* NTP_SERVER = "pool.ntp.org";
const char* TZ_INFO = "EST5EDT";
tm timeinfo;
time_t now;  // time_t is a type of variable, like 1707600003 with the last digit advancing each sec.
const char* PARAM_INPUT_1 = "alarmInputFromApp";

// Create AsyncWebServer object on port 80
AsyncWebServer server(80);


String getValue(String data, char separator, int desiredValue) {
  int maxIndex = data.length() - 1;
  /* Determining the maximum index of the individual characters in the string by looking at the length of the string.
  So if the data represents the time '10:20' then the data.length() = 5 because there are 5 characters in the data string.
  The maximum index then is 4 (again because indexes start counting at 0)
  */

  int valueBeingParsed = 0;
  // Declaring a variable to indicate which value is being concatenated below, we can use this to compare to desiredValue variable.
  // Resetting valueBeingParsed to zero so each time we start using this function it isn't an older value from a previous time

  String chunkVal = "";
  // declaring a temporary working string, called chunkVal, to allow us to parse through data[]. We are resetting chunkVal to zero before using it


  for (int i = 0; i <= maxIndex && valueBeingParsed <= desiredValue; i++) {
    chunkVal.concat(data[i]);
    /* concatenate (concat) is a function that adds one character at a time. So if the data is 10:20, then for this above line, first character, which is
    data[0] = '1' will be concatenated to chunkVal, and the code will continue down from here. When the ESP32 comes to this point again,
    because we are in a 'for loop' the next character of data[i] is data[1] which is '0' will be concatenated, so chunkVal will equal '10', and
    so on...
    */

    if (data[i] == separator) {
      valueBeingParsed = valueBeingParsed + 1;
      // If the character just concatenated is the delimiter then increment valueBeingParsed to indicate we are about to move on to the next
      // value in the data string


      if (valueBeingParsed > desiredValue) {
        chunkVal.trim();
        // if valueBeingParsed has is now greater than the desiredValue, then we are about to move past the value in the data
        // string that we want. So remove the separator character ':' with trim()
        // For example as we go through this for loop, chunkVal is '1', then we concatenated '0' on to that to make '10:'. Then we concatenated ':' but
        // since ':' is the 'separator' character we incremented valueBeingParsed (from the previous if statement above) and now with trim() we remove ':' from
        //chunkVal which is now '10'

        return chunkVal;
        // return means that this code will now give whatever value is in chunkVal as it exits this function
      }
      // if you are here in the code then you have concatenated to the point of the separator character, but valueBeingParsed was not greater than the desiredValue
      // (for example desiredValue = 1 and balueBeingParsed = 1). In our example, chunkVal = '10:' at his point

      chunkVal = "";
      /* In the line above we reset chunkVal. This is because we have concatenated the data one character at a time to chuckVal and at
      this point we have reached the separator character. However, according to desiredValue, we do not want the number in chunkVal
      at the moment (if desiredValue is 1, then we want the minutes, but chunkVal is currently '10:' for example). So, chunkVal is deleted in the above line and
      we will continue concatenating the data string to it as we proceed from here. In our example where data[] started as 10:20, then as we continue through in the 'for loop'
      from this point we will end up concatenating '2' to chunkVal, then '0' at which point chunkVal will equal '20' and int i will have reached maxIndex.
      Since i == maxIndex, the ESP32 will exit the 'for loop'
      */

    } else if ((desiredValue == maxIndex) && (valueBeingParsed < desiredValue)) {
      /* the ESP32 is here only if the concatenating has reached the end of the string (maybe because it was a partial data string?) but valueBeingParsed
       has not reached the desiredValue yet. So in the lines below, chunkVal is deleted and the blank chunkVal is returned because the value being sought
       was not avaialbe in the data string
       */

      chunkVal = "";
      return chunkVal;
    }  // This is the end of the 'for loop'
  }
  return chunkVal;  // The ESP32 at this point in the code after exiting the 'for loop', and now the current data in chunkVal is returned. From our example of '10:20", then '20' is returned
}


// This function gets the time from the Network Time Protocol from the WWW to ge the current time for our clock
bool getNTPtime(int sec) {
  {
    uint32_t start = millis();
    do {
      time(&now);
      localtime_r(&now, &timeinfo);
      delay(10);

    } while (((millis() - start) <= (1000 * sec)) && (timeinfo.tm_year < (2016 - 1900)));
    if (timeinfo.tm_year <= (2016 - 1900)) return false;  // the NTP call was not successful
  }
  return true;
}


// This function prints real time clock on serial monitor (used for testing)
void showTime(tm localTime) {
  Serial.println('\n');
  Serial.print(dayOfTheWeek(timeinfo));
  Serial.print(", ");
  Serial.print(localTime.tm_hour);
  Serial.print(':');
  Serial.print(localTime.tm_min);
  Serial.print(':');
  Serial.println(localTime.tm_sec);
}

// Get the Hour
String Hour(tm localTime) {
  int t = localTime.tm_hour;
  if (isnan(t)) {
    return "--";
  } else {
    return String(t);
  }
}

// Get the Minutes
String Minutes(tm localTime) {
  int t = localTime.tm_min;
  if (isnan(t)) {
    return "--";
  } else {
    if (t < 10) {
      return "0" + String(t);
    } else {
      return String(t);
    }
  }
}

// Get the Seconds
String Seconds(tm localTime) {
  int t = localTime.tm_sec;
  if (isnan(t)) {
    return "--";
  } else {
    if (t < 10) {
      return "0" + String(t);
    } else {
      return String(t);
    }
  }
}

// Get the day of the week
// https://mikaelpatel.github.io/Arduino-RTC/d8/d5a/structtm.html
String dayOfTheWeek(tm localTime) {
  DaY = localTime.tm_wday;
  char wday[10] = { 0 };
  if (isnan(DaY)) {
    return "--";
  } else {
    switch (DaY) {
      case 0:
        strcpy(wday, "SUNDAY");
        break;
      case 1:
        strcpy(wday, "MONDAY");
        break;
      case 2:
        strcpy(wday, "TUESDAY");
        break;
      case 3:
        strcpy(wday, "WEDNESDAY");
        break;
      case 4:
        strcpy(wday, "THURSDAY");
        break;
      case 5:
        strcpy(wday, "FRIDAY");
        break;
      case 6:
        strcpy(wday, "SATURDAY");
        break;
    }
    return String(wday);
  }
}

// Get the Months
String month(tm localTime) {
  int t = localTime.tm_mon;
  char mon[10] = { 0 };
  if (isnan(t)) {
    return "--";
  } else {
    switch (t) {
      case 0:
        strcpy(mon, "January");  // starts counting from zero
        break;
      case 1:
        strcpy(mon, "February");
        break;
      case 2:
        strcpy(mon, "March");
        break;
      case 3:
        strcpy(mon, "April");
        break;
      case 4:
        strcpy(mon, "May");
        break;
      case 5:
        strcpy(mon, "June");
        break;
      case 6:
        strcpy(mon, "July");
        break;
      case 7:
        strcpy(mon, "August");
        break;
      case 8:
        strcpy(mon, "September");
        break;
      case 9:
        strcpy(mon, "October");
        break;
      case 10:
        strcpy(mon, "November");
        break;
      case 11:
        strcpy(mon, "December");
        break;
    }
    return String(mon);
  }
}

// Get the Day of the Month
String dayOfTheMonth(tm localTime) {
  int t = localTime.tm_mday;
  if (isnan(t)) {
    return "--";
  } else {
    return String(t);
  }
}

// ******************************************************************** HTML CSS CODE STARTS HERE *************************************************************************************** //
const char index_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML><html>
<head>
  <title>App for ESP32 Alarm Clock</title>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <style>

body { font-family: arial black; padding-top: 10px; text-align: center; align-items: center; background-color: #02070d;}

.center1 {
  height: 75px; 
  width: 480px; 
  margin: auto;
  display: flex;
  justify-content: center;
  align-items: center;
}
.pic{
  width: 477px;
  height: 75px;
  margin: auto;
 }

.text1 {
  position: absolute;
  font-size: 50px;
  font-weight: bold;
  color: LightSkyBlue;
  z-index: -1;
}

.text2 {
  position: absolute;
  font-size: 50px;
  font-weight: bold;
  color: transparent;
  z-index: 1;
  -webkit-text-stroke: 3px #230414;"
}

.container1 {
  translate: 150px 40px; 
  display: flex;
  margin:auto;
  width: 100px;
  height: 75px;
  align-items: center;
  justify-content: center;
  text-align: center;
  font-family: arial;
  font-size: small;
}


input[type=text], select {
  display: inline-block;
  border: none;
  border-radius: 4px;
  box-sizing: border-box;
  z-index: 4;
  background-color: DarkBlue;
  color: white;
}


input[type=submit] {
  translate: 0px 10px;
  cursor: pointer;
  padding: 5px 5px;
  text-align: center;
  outline: none;
  color: White; 
  border: none;
  background-color: DarkBlue;
  border-radius: 5px;
  box-shadow: 0 6px #999;
  cursor: pointer;
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -khtml-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
  -webkit-tap-highlight-color: rgba(0,0,0,0);
  z-index: 4;
}

.center2 {
  display: flex;
  margin: 60px auto;
  width: 300px;
  height: 150px;
  align-items: center;
  justify-content: center;
  text-align: center;
  color: LightSkyBlue; 
}

.glass{
 z-index: 2;
  display: flex;
  margin: 120px auto;
  width: 300px;
  height: 150px;
  background-color: rgba(255,255,255,0.08);
  border-width: 1px;
  border-style: solid;
  border-color: rgba(255,255,255,0.06);
  box-shadow: 0 25px 30px rgba(0,0,0,0.5);
  backdrop-filter: blur(10px);
  align-items: center;
  justify-content: center;
  text-align: center;
  font-size: xxx-large;
  font-family: arial black;
  color: White; /* for text */
  -webkit-text-stroke: 1px #230414;
}

.shapeBlue{
  z-index: -1;
  position: absolute;
  translate: 150px -100px;
  width: 180px;
  height: 180px;
  border-radius: 90px;
  background-image: linear-gradient(to bottom right, RoyalBlue, Indigo); 
  display: flex;
  align-content: center;
}

.shapeRed{
  z-index: -1;
  position: absolute;
  translate: -150px 100px;
  width: 180px;
  height: 180px;
  border-radius: 90px;
  background-image: linear-gradient(to bottom right, Maroon, Tomato, Maroon); 
  display: flex;
  flex-direction: column;
  align-content: center;
  align-content: space-between;
}

.container2 {
  translate: -150px -72px; 
  display: flex;
  margin:auto;
  width: 100px;
  height: 75px;
  align-items: center;
  justify-content: center;
  text-align: center;
  font-size: small;
  font-family: Arial;
}

 .button {
   padding: 5px 5px;
   text-align: center;
   outline: none;
   color: White; 
   border: none;
   background-color: #942901;
   border-radius: 5px;
   box-shadow: 0 6px #999;
   cursor: pointer;
   -webkit-touch-callout: none;
   -webkit-user-select: none;
   -khtml-user-select: none;
   -moz-user-select: none;
   -ms-user-select: none;
   user-select: none;
   -webkit-tap-highlight-color: rgba(0,0,0,0);
   z-index: 4;
}
.button2 {
  position: absolute; 
  translate: 0px 45px; 
  padding: 5px 5px;
  text-align: center;
  outline: none;
  color: White; 
  border: none;
  background-color: #942901;
  border-radius: 5px;
  box-shadow: 0 6px #999;
  cursor: pointer;
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -khtml-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
  -webkit-tap-highlight-color: rgba(0,0,0,0);
  z-index: 4;
}

button:hover {background-color: #5c1a01}
  button:active {
  background-color: #5c1a01;
  box-shadow: 0 4px #666;
  transform: translateY(2px);
  }
button2:hover {background-color: #5c1a01}
button2:active {
  background-color: #5c1a01;
  box-shadow: 0 4px #666;
  transform: translateY(2px);
   }
input:hover[type=submit] {background-color: #00003b}
input:active[type=submit] {
  background-color: #00003b;
  box-shadow: 0 4px #666;
  transform: translateY(2px);
}

  </style>
</head>
<body>

<br/>
<div class="center1">
<img class="pic" src="image"></img>
  <span class="text1">%DAY%,&nbsp%MONTH%&nbsp%DATE%</span>
  <span class="text2">%DAY%,&nbsp%MONTH%&nbsp%DATE%</span>
</div>

<div class="container1">
      <form action="/get">
      <label > </label>
      <input type="time" id="inputText" name="alarmInputFromApp" size="400" onchange="timers(this.value)" style="background-color: LightSkyBlue;"/>
      <input type="submit" value="ACTIVATE ALARM" id="activate">
      </form>
</div>      

<div class="center2">
    <div class="glass">
      <span id="HOURS">%HOURS%</span> :
      <span id="MINUTES">%MINUTES%</span> :
      <span id="SECONDS">%SECONDS%</span>
    </div>
    <div class="shapeBlue"></div>    
    <div class="shapeRed"></div>    
</div>

<div class="container2">    
       <button class="button" onmousedown="toggleCheckbox('off');" ontouchstart="toggleCheckbox('off');">ALARM OFF</buton>
       <button class="button2" onmousedown="toggleCheckbox('disable');" ontouchstart="toggleCheckbox('disable');">DISABLE ALL ALARMS</button>
</div>

<script>

// ***** Function for the online buttons - Only need one function for BOTH buttons *****
   function toggleCheckbox(x) {
     var xhr = new XMLHttpRequest();
     xhr.open("GET", "/" + x, true);
     xhr.send();
     this.value = "";
   }

// ****In BODY instead of %DAY% put a value getting from the function **** //
setInterval(function ( ) {
  var xhttp = new XMLHttpRequest();
  // Serial.println(xhttp);
  xhttp.onreadystatechange = function() {
    if (this.readyState == 4 && this.status == 200) {
      document.getElementById("DAY").innerHTML = this.responseText;
      }
  };
  xhttp.open("GET", "/DAY", true);
  xhttp.send();
}, 1000 ) ;


// ****In BODY instead of %MONTH% put a value getting from the function **** //
setInterval(function ( ) {
  var xhttp = new XMLHttpRequest();
  // Serial.println(xhttp);
  xhttp.onreadystatechange = function() {
    if (this.readyState == 4 && this.status == 200) {
      document.getElementById("MONTH").innerHTML = this.responseText;
      }
  };
  xhttp.open("GET", "/MONTH", true);
  xhttp.send();
}, 1000 ) ;
// TODO: Change interval, only need it when DAY=1

// ****In BODY instead of %DATE% put a value getting from the function **** //
setInterval(function ( ) {
  var xhttp = new XMLHttpRequest();
  // Serial.println(xhttp);
  xhttp.onreadystatechange = function() {
    if (this.readyState == 4 && this.status == 200) {
      document.getElementById("DATE").innerHTML = this.responseText;
      }
  };
  xhttp.open("GET", "/DATE", true);
  xhttp.send();
}, 1000 ) ;
// TODO: COMBINE WITH DAY and change interval, only need to do it when hour=1


// ****In BODY instead of %HOURS% put a value getting from the function **** //
setInterval(function ( ) {
  var xhttp = new XMLHttpRequest();
  // Serial.println(xhttp);
  xhttp.onreadystatechange = function() {
    if (this.readyState == 4 && this.status == 200) {
      document.getElementById("HOURS").innerHTML = this.responseText;
      }
  };
  xhttp.open("GET", "/HOURS", true);
  xhttp.send();
}, 1000 ) ;

// ****In BODY instead of %MINUTES% put a value getting from the function **** //
setInterval(function ( ) {
  var xhttp = new XMLHttpRequest();
  xhttp.onreadystatechange = function() {
    if (this.readyState == 4 && this.status == 200) {
      document.getElementById("MINUTES").innerHTML = this.responseText;
    }
  };
  xhttp.open("GET", "/MINUTES", true);
  xhttp.send();
}, 500 ) ;

// ****In BODY instead of %SECONDS% put a value getting from the function **** //
setInterval(function ( ) {
  var xhttp = new XMLHttpRequest();
  // Serial.print("Seconds: );
  // Serial.println(xhttp);
  xhttp.onreadystatechange = function() {
    if (this.readyState == 4 && this.status == 200) {
      document.getElementById("SECONDS").innerHTML = this.responseText;
    }
  };
  xhttp.open("GET", "/SECONDS", true);
  xhttp.send();
}, 10 ) ;


</script>
</body>
</html>
)rawliteral";
// ******************************************************************** HTML CSS CODE ENDS HERE *************************************************************************************** //

String processor(const String& var) {
  if (var == "HOURS") {
    return Hour(timeinfo);
  } else if (var == "MINUTES") {
    return Minutes(timeinfo);
  } else if (var == "SECONDS") {
    return Seconds(timeinfo);
  } else if (var == "DAY") {
    return dayOfTheWeek(timeinfo);
  } else if (var == "MONTH") {
    return month(timeinfo);
  } else if (var == "DATE") {
    return dayOfTheMonth(timeinfo);
  }
  return String();
}


void checkAlarmTime(int hour, int min) {
  if (alarmState == alarmSet && dailyMode==0) {
    if (DaY == 0 || DaY == 6) return;
  }
  if (Hour(timeinfo).toInt() == hour) {     
    if (Minutes(timeinfo).toInt() == min) {  
      buzzerFlag = 1;
      valuesUpdated=0;
      buzzerStartMillis = millis();  
      if (alarmState == alarmSet) alarmState=alarmSounding;
    }
  }
 }


void buzzBuzz() {

  ledcWrite(0, 255);  
  if (toggleTone == 1) {
    ledcWriteTone(0, 800);  
  } else {
    ledcWriteTone(0, 262);  
  }
  toggleTone = !toggleTone;
  toggleTime = (millis() + toneDuration); 
  digitalWrite(LEDpin,toggleTone); 
}


void setSnooze(int delay) {
  buzzerFlag = 0;
  // snoozeMinute = (localTime.tm_min) + delay;  
  snoozeMinute = (Minutes(timeinfo).toInt() + delay);  
  if (snoozeMinute > 59) {
    snoozeMinute = snoozeMinute - 60;
    snoozeHour = (Hour(timeinfo).toInt() + 1);
  } else {
    snoozeHour = Hour(timeinfo).toInt();
  }
}


void checkAmbientLight() {
  digitalWrite(LEDpin,LOW);
  ambientLight = analogRead(LDRpin);
  for (int i = 0; i < 3; i++) {  
    ambientLight = ambientLight + analogRead(LDRpin);
    ambientLight = ambientLight / 2;
  }
  lightThreshold = (ambientLight +200);
  if (ambientLight >= ambientTooBright) ambientLight = (ambientLight - 10);  // so bright that it is unlikely to get brighter
  ambientChecked = true;
}


void resetForTomorrow() {
  buzzerFlag = 0;
  ambientChecked = false;     
  toneDuration = 1200;         
  prevLEDtime = 0;
  dailyMode = digitalRead(dailyModePin);
  alarmState = alarmSet;      
}


void playSound(int Tone, int Duration) {
  //ledcSetup(0, Tone, 12); 
  ledcWriteTone(0, Tone);
  ledcWrite(0, 50);  
  unsigned long delayTime = millis() + Duration;
  while(millis() < delayTime){}
  ledcWrite(0, 0);  
}

void confirmSound(int tone1, int duration1, int tone2, int duration2) {
  // Winning sound: 262Hz, 300msec, 659Hz, 400msec
  // Error sound: 2500Hz, 200mSec,200H, 700 msec
  playSound(tone1, duration1);
  delay(100); // even though the each tone should last for the specified duration it gets cut off
  playSound(tone2, duration2);
  delay(100);
}

void flashLED (long duration) {
  if (millis() >= prevLEDtime) {
    digitalWrite(LEDpin,!digitalRead(LEDpin));
    prevLEDtime=(millis() + duration);
  }
}

// ********************  Code for the shift registers and 7-seg display **********************
void parseTimeData(int Hour, int Minute) {
  byte tempNumbers[] = {0, 0, 0, 0};
  // get the two digits for hour
  int tempNumber = Hour;
  tempNumbers[3] = (tempNumber / 10);
  tempNumbers[2] = (tempNumber % 10);

  // get the two digits for minutes
  tempNumber = Minute;
  tempNumbers[1] = (tempNumber / 10);
  tempNumbers[0] = (tempNumber % 10);

// convert each number to a pattern for the display
xSemaphoreTake(baton, portMAX_DELAY); // Uses 'mutex' (mutually exclusive') semaphore to prevent task1 from reading the time variables while they are updated
for (int j=0; j<4; j++) {
  switch (tempNumbers[j]) {
    case 0: timeValues[j] = 0b00111111; // 0; 255-128 = 127
      break;
    case 1: timeValues[j] = 0b00000110; // 1; =6
      break;
    case 2: timeValues[j] = 0b01011011; // 2; 64+16+8+2+1 = 91
      break;
    case 3: timeValues[j] = 0b01001111; // 3; 64+16+4+2+1 = 87
      break;
    case 4: timeValues[j] = 0b01100110; // 4; 64+32+4+2 = 102
      break;
    case 5: timeValues[j] = 0b01101101; // 5; 64+32+8+4+1 = 109
      break;
    case 6: timeValues[j] = 0b01111101; // 6; 255-130 = 125
      break;
    case 7: timeValues[j] = 0b00000111; // 7; 4+2+1 = 7
      break;
    case 8: timeValues[j] = 0b01111111; // 8; 127
      break;
    case 9: timeValues[j] = 0b01100111; // 9; 255-128-16-8 = 103
      break;
    default: timeValues[j] = 0b00000000; // clear =0
  }
  if (j==2) timeValues[2]=timeValues[2]+0b10000000; // to add the colon
  if (j==3 && (timeValues[3] == 0b00111111)) timeValues[3]=0b00000000; // clears the digit if it is '0'
  xSemaphoreGive(baton); // releases the baton so that Task can read the time variables
}
// now the array has the pattern for lighting up the LEDs ('A' to 'G') of the 7-segment display, and not the number itself
}

byte getDigit(byte number)
{
  byte digit=0b00000000; // without giving an intial value here, an error is generated for 'return digit; below when compilied...
  switch (number) {
    case 1: digit=(0b00000001); 
      break;
    case 2: digit=(0b00000010); 
      break;
    case 3: digit=(0b00000100); 
      break;
    case 4: digit=(0b00001000); 
      break;
  }
  return digit;
}


void displayOneDigit(byte n){
   digitalWrite(latchPin, LOW);
   shiftOut(dataPin, clockPin, MSBFIRST, getDigit(n));  
   shiftOut(dataPin, clockPin, MSBFIRST, timeValues[n-1]); 
   digitalWrite(latchPin, HIGH);
  }

// send the signal to the shift registers to light up the 7-segment display    
void codeForTask0( void * parameter)
{
const TickType_t xDelay0 = delayTime / portTICK_PERIOD_MS; 

for(indexDigit=0; indexDigit <= 6;indexDigit++) {
xSemaphoreTake(baton, portMAX_DELAY); // Take the baton: the 'semaphore' is a way to prevent the two tasks that use the same resource (ie, varible) from crashing together
  displayOneDigit(indexDigit);
xSemaphoreGive(baton); // Give the baton so that now the other task can access the shared variable 
  if (indexDigit==5) indexDigit=0;
vTaskDelay(xDelay0);
} 
}

void showAlarm() {
  parseTimeData(alarmHour, alarmMinute);
  bool toggle=0;
  unsigned long endTime= (millis() + 4000);
  while (millis() <= endTime){
    digitalWrite(LEDpin,HIGH);
    delay(75);
    digitalWrite(LEDpin,LOW);
    delay(175);
    if (toggle==1) delay(500);
    toggle=!toggle;
  }
 }


void setup() {
  Serial.begin(115200);
  pinMode(buttonPin, INPUT_PULLUP);  
  ledcSetup(0, 800, 12);        
  ledcAttachPin(buzzerPin, 0);  
  pinMode(LEDpin, OUTPUT);
  digitalWrite(LEDpin, LOW);  
  pinMode(latchPin, OUTPUT); 
  pinMode(dataPin, OUTPUT); 
  pinMode(clockPin, OUTPUT); 
  pinMode(dailyModePin,INPUT_PULLUP); 


  // Initialize SPIFFS
  if(!SPIFFS.begin(true)){
    Serial.println("An Error has occurred while mounting SPIFFS");
    return;
  }

  // Connect to Wi-Fi
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to WiFi..");
  }

  // Print ESP32 Local IP Address
  Serial.println(WiFi.localIP());  

  // Time Configuration
  configTime(0, 0, NTP_SERVER);  
  setenv("TZ", TZ_INFO, 1);  

  if (getNTPtime(10)) {  
      } else {
     ESP.restart();
  }
  showTime(timeinfo);  

  server.on("/", HTTP_GET, [](AsyncWebServerRequest* request) {
    request->send_P(200, "text/html", index_html, processor);
  });

  server.on("/HOURS", HTTP_GET, [](AsyncWebServerRequest* request) {  
    request->send_P(200, "text/plain", Hour(timeinfo).c_str());
  });
  server.on("/MINUTES", HTTP_GET, [](AsyncWebServerRequest* request) {
    request->send_P(200, "text/plain", Minutes(timeinfo).c_str());
  });
  server.on("/SECONDS", HTTP_GET, [](AsyncWebServerRequest* request) {
    request->send_P(200, "text/plain", Seconds(timeinfo).c_str());
  });
  server.on("/DAY", HTTP_GET, [](AsyncWebServerRequest* request) {
    request->send_P(200, "text/plain", dayOfTheWeek(timeinfo).c_str());
  });
  server.on("/MONTH", HTTP_GET, [](AsyncWebServerRequest* request) {
    request->send_P(200, "text/plain", month(timeinfo).c_str());
  });
  server.on("/DATE", HTTP_GET, [](AsyncWebServerRequest* request) {
    request->send_P(200, "text/plain", dayOfTheMonth(timeinfo).c_str());
  });

//  **** This is used to tell the ESP32 to get the image file from the file that was uploaded on to the ESP32 with SPIFFS 
  server.on("/image", HTTP_GET, [](AsyncWebServerRequest* request) {
     request->send(SPIFFS, "/image.png", "image/png");
  });

  // Receive an HTTP GET request for the "Alarm off"
  server.on("/off", HTTP_GET, [](AsyncWebServerRequest* request) {
    request->send(200, "text/plain", "ok");
    if ((alarmState >= alarmSounding) && (alarmState < alarmOff)) {
      // Remember enum states are like numbers 0-5 in this case
      alarmState = alarmOff;  
      buzzerFlag = 0;
    } else {
      confirmSound(2500, 200, 200, 500);  // Error sound: 2500Hz, 200mSec,200H, 700 msec
    }
  });

  // Receive an HTTP GET request for the "disable alarm"
  server.on("/disable", HTTP_GET, [](AsyncWebServerRequest* request) {
    request->send(200, "text/plain", "ok");
    alarmState = noAlarm;
    valuesUpdated=0;
    buzzerFlag = 0;
    confirmSound(262, 300, 659, 200);  // Winning sound: 262Hz, 300msec, 659Hz, 400msec
  });

  // ******************************************************* HERE WE GET THE INPUT FROM THE WEBSITE (ALARM SET) ********************************************************************** //
  server.on("/get", HTTP_GET, [](AsyncWebServerRequest* request) {
    request->send_P(200, "text/html", index_html, processor);
    String inputMessage;
    if (request->hasParam(PARAM_INPUT_1)) {
      inputMessage = request->getParam(PARAM_INPUT_1)->value();  
      if (inputMessage == "") {
        confirmSound(2500, 200, 200, 500);  // Error sound: 2500Hz, 200mSec,200H, 700 msec
      } else {
         confirmSound(262, 300, 659, 200);  // Winning sound: 262Hz, 300msec, 659Hz, 400msec
         alarmHour =(getValue(inputMessage, ':', 0).toInt());
         alarmMinute = (getValue(inputMessage, ':', 1).toInt());
         alarmState = alarmSet;
        resetForTomorrow();
      }
    } else {
      confirmSound(2500, 200, 200, 500);  // Error sound: 2500Hz, 200mSec,200H, 700 msec
    }
    // Serial.println(inputMessage);
  });
  
// ************************************************************************************************************************************************************************** //

// Start server
  server.begin();

  
baton = xSemaphoreCreateMutex();

xTaskCreatePinnedToCore(
  codeForTask0,   /* Task function */
  "displayTimeTask0",    /* name of task */
  1000,     /* Stack size of task */
  NULL,     /*parameter of the task */
  1,      /* priority of the task */
  &displayTimeTask0,     /* Task handle to keep track of created task */
  1);     /* Indicates which core */

}  // end of setup()


void loop() {

getNTPtime(10);  
parseTimeData(Hour(timeinfo).toInt(), Minutes(timeinfo).toInt()); 
delay(200);

  switch (alarmState) {
    
    case noAlarm:  
    if (valuesUpdated == 0) {
      resetForTomorrow();
      digitalWrite(LEDpin,LOW);
      alarmState=noAlarm;
      valuesUpdated=1;
    }
      break;

    case alarmSet:  
        checkAlarmTime(alarmHour, alarmMinute);
        if (dailyMode) {digitalWrite(LEDpin,HIGH);} else {flashLED(1000);}
        if (digitalRead(buttonPin)==0) showAlarm();
       break;

    case alarmSounding:  
      if (ambientChecked == false) checkAmbientLight();
      if (digitalRead(buttonPin) == 0) {
        setSnooze(10);
        alarmState = firstSnooze;
      }
      break;

    case firstSnooze:  
      if (buzzerFlag == 0) {
        checkAlarmTime(snoozeHour, snoozeMinute);
        flashLED(100);
      } else {
        if ((digitalRead(buttonPin) == 0) && (analogRead(LDRpin) >= (lightThreshold))) {
          setSnooze(1);  // Snoozes only until the the clock changes to the next min (i.e., 1 min max)
          alarmState = overSleeping;
        }
      }
      break;

    case overSleeping:
      if (buzzerFlag == 0) {
        checkAlarmTime(snoozeHour, snoozeMinute);
        flashLED(100);
      } else {
        toneDuration = random(10, 500);  //randomizing the duration between switching tones
        if ((digitalRead(buttonPin) == 0) && (analogRead(LDRpin) >= (lightThreshold))) {
          setSnooze(1);
        }
      }
      if (analogRead(LDRpin) < lightThreshold) buzzerFlag = 1;
      break;

    case alarmOff:        
    if (alarmMinute == Minutes(timeinfo).toInt()) {
        if (alarmHour == Hour(timeinfo).toInt()) {

        }
      } else if (analogRead(LDRpin) > lightThreshold) {
        resetForTomorrow();
      }

      if (analogRead(LDRpin) < lightThreshold) {
        alarmState = overSleeping;
        buzzerFlag = 1;
        buzzerStartMillis = millis();
      }
      break;
  }  // end of the switch case


  if (buzzerFlag == 1) {
    // Here is where the alarm is sounded regardless of the alarmState
    if (millis() > toggleTime) buzzBuzz();
    if ((millis() - buzzerStartMillis) > buzzerOverflow) {
      // Checking if alarm has not been snoozed in over 2hours - turn off alarm so it doesn't sound for 24hrs
      resetForTomorrow();
    }
  } else {
    ledcWrite(0, 0);  //Setting PWM to 0 = "off"; (ledChannel, dutyCycle)
  }

}  // end of the loop
